//=============================================================================
// File: com.c
// Func: Serial Communication Functions
// Rem.: Receive Interrupt-driven, Transmit uses Polling
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// includes

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "gpio.h"
#include "usart.h"
#include "intc.h"
#include "pm_at32ap7000.h"
#include "error.h"
#include "global.h"
#include "com.h"

//-----------------------------------------------------------------------------
// definitions

#define STDIO_USART					(&AVR32_USART1)
#define STDIO_USART_RX_PIN			AVR32_USART1_RXD_0_PIN
#define STDIO_USART_RX_FUNCTION		AVR32_USART1_RXD_0_FUNCTION
#define STDIO_USART_TX_PIN			AVR32_USART1_TXD_0_PIN
#define STDIO_USART_TX_FUNCTION		AVR32_USART1_TXD_0_FUNCTION
#define STDIO_USART_IRQ				AVR32_USART1_IRQ
#define STDIO_USART_BAUDRATE		115200

#define MAX_LINES 10

#define	TERMINAL
//#define KEYBOARD

#define STATE0						0
#define STATE1						1
#define STATE2						2
#define STATE3						3

#ifdef TERMINAL
#define EXCEPTION					'\033'
#define ARROW_UP					'A'
#define ARROW_DOWN					'B'
#define ARROW_RIGHT					'C'
#define ARROW_LEFT					'D'
#define HOME						'1'
#define INSERT						'2'
#define DELETE						'3'
#define END							'4'
#define PAGE_UP						'5'
#define PAGE_DOWN					'6'
#else
#define EXCEPTION					0xE0
#define ARROW_UP					'H'
#define ARROW_DOWN					'P'
#define ARROW_RIGHT					'M'
#define ARROW_LEFT					'K'
#define HOME						'G'
#define INSERT						'R'
#define DELETE						'S'
#define END							'O'
#define PAGE_UP						'I'
#define PAGE_DOWN					'Q'
#endif

//-----------------------------------------------------------------------------
// global variables

//-----------------------------------------------------------------------------
// variables

static char						*kbLine;
static char						*bufLine;
static char						*lastLine[MAX_LINES];

static int						bufIndex;
static int						lineWritePtr;
static int						lineReadPtr;


static char						buffer[BUFFERSIZE];		// SCI0 Receive Buffer

volatile int	readIndex;				// RX Buffer Read Index
volatile int	writeIndex;				// RX Buffer Write Index
//volatile static unsigned short	readIndex;				// RX Buffer Read Index
//volatile static unsigned short	writeIndex;				// RX Buffer Write Index
volatile static char			stopFlag;
volatile static char			xoffFlag;
//static char						escFlag;

extern char	gEcho;


//-----------------------------------------------------------------------------
// functions

static const gpio_map_t	usartGPIOMap =
{
	{STDIO_USART_RX_PIN, STDIO_USART_RX_FUNCTION},
	{STDIO_USART_TX_PIN, STDIO_USART_TX_FUNCTION}
};

// USART options.
static const usart_options_t usartOptions =
{
	.baudrate     = STDIO_USART_BAUDRATE,
	.charlength   = 8,
	.paritytype   = USART_NO_PARITY,
	.stopbits     = USART_1_STOPBIT,
	.channelmode  = USART_NORMAL_CHMODE
};

//-----------------------------------------------------------------------------
/*! \brief The USART interrupt handler.
 *
 * \note The `__attribute__((__interrupt__))'
 * attribute is used to manage the `rete' instruction.
 */
__attribute__((__interrupt__))
static void UsartIntHandler(void)
{
	int 			c, error, test;
//	unsigned short	test;
//	volatile int	cnt = 10;

	// In the code line below, the interrupt priority level does not need to be
	// explicitly masked as it is already because we are within the interrupt
	// handler.
	// The USART Rx interrupt flag is cleared by side effect when reading the
	// received character.
	// Waiting until the interrupt has actually been cleared is here useless as
	// the call to usart_write_char will take enough time for this before the
	// interrupt handler is leaved and the interrupt priority level is unmasked by
	// the CPU.
	error = usart_read_char( STDIO_USART, &c );
	if (error) return;
//	buffer[writeIndex++] = (char)c;
//	if (writeIndex >= BUFFERSIZE) writeIndex = 0;


	if ((char)c == XOFF)
	{
		xoffFlag = 1;
	}
	else if ((char)c == XON)
	{
		xoffFlag = 0;
	}
	else
	{
		test = writeIndex + 1;
		if(test >= BUFFERSIZE) test = 0;
		if(test == readIndex)
		{
			usart_write_char( STDIO_USART, (int)XOFF );
			stopFlag = 1;										// and flag it
		}
		else
		{
			buffer[writeIndex++] = (char)c;
			if (writeIndex >= BUFFERSIZE) writeIndex = 0;

			test = writeIndex - readIndex;
			if (test < 0) test += BUFFERSIZE;
			test = test - BUFFERSIZE + REMAINING;
			if (test > 0)
//			if ((GetBufferCnt()) >= (BUFFERSIZE - REMAINING))
			{
				usart_write_char( STDIO_USART, (int)XOFF );
				stopFlag = 1;										// and flag it
			}

		}

	}

//	while (--cnt);
}

//-----------------------------------------------------------------------------

void SetupUART( void )
{
	int		i;


	// Assign GPIO to USART.
	gpio_enable_module( usartGPIOMap, sizeof(usartGPIOMap) / sizeof(usartGPIOMap[0]) );

	// Initialize USART in RS232 mode.
	usart_init_rs232( STDIO_USART, &usartOptions, pm_read_module_freq_hz( PM_PBA_USART1 ) );

	// Register the USART interrupt handler to the interrupt controller.
	// UsartIntHandler is the interrupt handler to register.
	// STDIO_USART_IRQ is the IRQ of the interrupt handler to register.
	// AVR32_INTC_INT0 is the interrupt priority level to assign to the group of
	// this IRQ.
	// void INTC_register_interrupt(__int_handler handler, unsigned int irq, unsigned int int_lev);
	INTC_register_interrupt( &UsartIntHandler, STDIO_USART_IRQ, AVR32_INTC_INT0 );

	// Enable USART Rx interrupt.
	STDIO_USART->ier = AVR32_USART_IER_RXRDY_MASK;

	stopFlag = 0;
	xoffFlag = 0;
//	escFlag = STATE0;

	readIndex = 0;
	writeIndex = 0;

	bufIndex = 0;

	lineWritePtr = 0;
	lineReadPtr = 0;
	for (i = 0; i < MAX_LINES; i++)
	{
		lastLine[i] = NULL;
	}
	if((kbLine = malloc( sizeof( char ) * 1 )) == NULL ) return;
	kbLine[0] = '\0';
	bufLine = NULL;
}
//-----------------------------------------------------------------------------
/*
unsigned short GetBufferCnt( void ) 
{
	short cnt;

	cnt = writeIndex - readIndex;
	if (cnt < 0) cnt += BUFFERSIZE;
	return cnt;
}
*/
int GetBufferCnt( void )
{
	int cnt;

	cnt = writeIndex - readIndex;
	if (cnt < 0) cnt += BUFFERSIZE;
	return cnt;
}

//-----------------------------------------------------------------------------

char GetChar( void )
{
	char 	received;

	if (stopFlag)
	{
		if ((GetBufferCnt()) < (BUFFERSIZE - 2 * REMAINING))
		{
			PutChar( XON );	// enable sending
			stopFlag = 0;
		}
	}
	while(readIndex == writeIndex);		// if buffer empty wait for character
	received = buffer[readIndex++];
	if (readIndex >= BUFFERSIZE) readIndex = 0;
	return received;
}

//-----------------------------------------------------------------------------

void UngetChar( char c )
{
	if(readIndex == 0) readIndex = BUFFERSIZE;
	buffer[--readIndex] = c;
}


//-----------------------------------------------------------------------------

void PutChar( char c )
{
	while (xoffFlag);
	usart_putchar( STDIO_USART, (int)c );
}

//-----------------------------------------------------------------------------

void EchoChar( char c )
{
	if (gEcho)
	{
		while (xoffFlag);
		usart_putchar( STDIO_USART, (int)c );
	}
}

//-----------------------------------------------------------------------------
void Prints( const char *text )
{
	while (xoffFlag);
	usart_write_line( STDIO_USART, text );
}

//-----------------------------------------------------------------------------
void Printf( char *format, ... )
{
	static char		buffer[120 + 1];
	va_list			vArgs;

	va_start( vArgs, format );
	vsnprintf( (char *)buffer, sizeof( buffer ), (const char*)format, vArgs );
	va_end( vArgs );

	while (xoffFlag);
	usart_write_line( STDIO_USART, (char*)buffer );
}

//-----------------------------------------------------------------------------
void FlushReceiverBuffer( void )
{
	readIndex = writeIndex;
}

//-----------------------------------------------------------------------------
char *AllocateMemory( char *charPtr, int amount )
{
	char	*ptr = charPtr;

	if (ptr == NULL)
	{
		if((ptr = malloc( sizeof( char ) * (amount + 1) )) == NULL )
		{
			return NULL;
		}
	}
	else
	{
		if ((ptr = realloc( ptr, sizeof( char ) * (amount + 1) )) == NULL )
		{
			return NULL;
		}
	}
	return ptr;
}

/*****************************************************************
* Function:     ReadLine()
* Arguments:    none
* Returns:      a pointer to the buffer if successful, NULL else
******************************************************************/
extern char *ReadLine( int *error )
{
	unsigned char	c;
	int				index;
	int				len;
	static char		relChar[4];
	static char		escFlag = 0;


#ifdef TERMINAL
	if (GetBufferCnt() == 0) return NULL;
	c = GetChar();
#else
	c = getch();
#endif

	*error = 0;
	len = strlen( kbLine );
//	if (len >= MAX_LINE_LENGTH && c != '\r') return NULL;

/*
	Printf( "%02X ", c );
	if (c == '\r') Prints( "\n" );
	return NULL;
*/

	switch (escFlag)
	{
		case STATE0:
			switch (c)
			{
				case '\r':
					EchoChar( '\n' );

					if ((lastLine[lineWritePtr] = AllocateMemory( lastLine[lineWritePtr], len + 1 )) == NULL)
					{
						*error = NO_MEMORY;
						return NULL;
					}
					strcpy( lastLine[lineWritePtr], kbLine );

					if ((bufLine = AllocateMemory( bufLine, len + 1 )) == NULL)
					{
						*error = NO_MEMORY;
						return NULL;
					}
					strcpy( bufLine, kbLine );

					if (++lineWritePtr >= MAX_LINES) lineWritePtr =  0;
					lineReadPtr = lineWritePtr;

					if ((lastLine[lineWritePtr] = AllocateMemory( lastLine[lineWritePtr], len + 1 )) == NULL)
					{
						*error = NO_MEMORY;
						return NULL;
					}
					lastLine[lineWritePtr][0] = '\0';

					if ((kbLine = realloc( kbLine, sizeof( char ) * (len + 1) )) == NULL )
					{
						*error = NO_MEMORY;
						return NULL;
					}
					kbLine[0] = '\0';
					bufIndex = 0;
					return bufLine;

				case '\n':						// ignor line feeds
				case '\t':						//  and tabs
					break;

				case '\b':
					if ( bufIndex > 0 )
					{
						bufIndex--;
						index = bufIndex;
						EchoChar( '\b' );
						while (kbLine[index])
						{
							kbLine[index] = kbLine[index+1];
							if (kbLine[index])
							{
								EchoChar( kbLine[index] );
							}
							else
							{
								EchoChar( ' ' );
							}
							index++;
						}

						while (index > bufIndex)
						{
							EchoChar( '\b' );
							index--;
						}
					}
					break;

				case EXCEPTION:					// escape
#ifdef TERMINAL
					escFlag = STATE1;
#else
					escFlag = STATE2;
#endif
					relChar[0] = EXCEPTION;
					break;

				default:
					len++;
					if ((kbLine = realloc( kbLine, sizeof( char ) * (len + 1) )) == NULL )
					{
						*error = NO_MEMORY;
						return NULL;
					}
					while (len > bufIndex)
					{
						kbLine[len] = kbLine[len-1];
						len--;
					}
					kbLine[bufIndex++] = c;

					index = bufIndex - 1;
					while (kbLine[index])
					{
						EchoChar( kbLine[index++] );
					}
					while (index > bufIndex)
					{
						EchoChar( '\b' );
						index--;
					}
					break;
			}
		break;

		case STATE1:
			if (c == '[')
			{
				escFlag = STATE2;
				relChar[1] = '[';
			}
			else
			{
				escFlag = STATE0;
			}
			return NULL;

		case STATE2:
			switch (c)
			{
				case HOME:
				case INSERT:
				case DELETE:
				case END:
				case PAGE_UP:
				case PAGE_DOWN:
					escFlag = STATE3;
					relChar[2] = c;
#ifdef TERMINAL
					return NULL;
#else
					c = '~';
					break;
#endif

				case ARROW_UP:
				case ARROW_DOWN:
					if (lineReadPtr == lineWritePtr)
					{
						if ((lastLine[lineWritePtr] = AllocateMemory( lastLine[lineWritePtr], len + 1 )) == NULL)
						{
							*error = NO_MEMORY;
							return NULL;
						}
						strcpy( lastLine[lineWritePtr], kbLine );
					}

					do
					{
						if (c == ARROW_UP)	// arrow key up
						{
							if (--lineReadPtr < 0) lineReadPtr = MAX_LINES - 1;
						}
						else				// arrow key down
						{
							if (++lineReadPtr >= MAX_LINES) lineReadPtr = 0;
						}
						if (lastLine[lineReadPtr] != NULL && lastLine[lineReadPtr][0] != '\0') break;
					} while (lineReadPtr != lineWritePtr);

					while (bufIndex)						// delete the written line
					{
						EchoChar( '\b' );
						EchoChar( ' ' );
						EchoChar( '\b' );
						bufIndex--;
					}

					len = strlen( lastLine[lineReadPtr] );
					if ((kbLine = realloc( kbLine, sizeof( char ) * (len + 1) )) == NULL )
					{
						*error = NO_MEMORY;
						return NULL;
					}
					while ((kbLine[bufIndex] = lastLine[lineReadPtr][bufIndex]))
					{
						EchoChar( kbLine[bufIndex++] );
					}
					escFlag = STATE0;
					break;

				case ARROW_RIGHT:
					if (bufIndex < len)
					{
						EchoChar( kbLine[bufIndex++] );

					}
					escFlag = STATE0;
					break;

				case ARROW_LEFT:
					if (bufIndex)
					{
						EchoChar( '\b' );
						bufIndex--;
					}
					escFlag = STATE0;
					break;
			}
//			break;

		case STATE3:
			if (c == '~')
			{
				if (relChar[2] == DELETE)			// Delete
				{
					index = bufIndex;
					while (kbLine[index])
					{
						kbLine[index] = kbLine[index+1];
						if (kbLine[index])
						{
							EchoChar( kbLine[index] );
						}
						else
						{
							EchoChar( ' ' );
						}
						index++;
					}

					while (index > bufIndex)
					{
						EchoChar( '\b' );
						index--;
					}
				}
				else if (relChar[2] == HOME)		// Home
				{
					while (bufIndex)
					{
						EchoChar( '\b' );
						bufIndex--;
					}

				}
				else if (relChar[2] == END)		// End
				{
					while (bufIndex < len)
					{
						EchoChar( kbLine[bufIndex++] );

					}

				}
			}
			escFlag = STATE0;
			break;

		default:
			escFlag = STATE0;
			break;
	}

	return NULL;
}

/*****************************************************************
* Function:     ReadLine()
* Arguments:    none
* Returns:      a pointer to the buffer if successful, NULL else
******************************************************************/
/*
extern char *ReadLine( void )
{
	unsigned char	c;
	int				index;
	size_t			len;


	if (GetBufferCnt() == 0) return NULL;
	c = GetChar();
	len = strlen( kbLine );
	if (len >= MAX_LINE_LENGTH && c != '\r') return NULL;

//	Printf( "%02X ", c );
//	return NULL;

	switch (escFlag)
	{
		case 0:
			switch (c)
			{
				case '\r':
					PutChar( '\n' );

					bufIndex = 0;
					strcpy( lastLine[lineWritePtr], kbLine );
					strcpy( bufLine, kbLine );
					if (++lineWritePtr >= MAX_LINES) lineWritePtr =  0;
					lineReadPtr = lineWritePtr;
					lastLine[lineWritePtr][0] = '\0';
					kbLine[0] = '\0';
					return bufLine;

				case '\n':						// ignor line feeds
					break;

				case '[':

				case '\b':
					if ( bufIndex > 0 )
					{
						bufIndex--;
						index = bufIndex;
						PutChar( '\b' );
						while (kbLine[index])
						{
							kbLine[index] = kbLine[index+1];
							if (kbLine[index])
							{
								PutChar( kbLine[index] );
							}
							else
							{
								PutChar( ' ' );
							}
							index++;
						}

						while (index > bufIndex)
						{
							PutChar( '\b' );
							index--;
						}
					}
					break;

				case '\033':					// escape
					escFlag = 1;
					break;

				default:
					len++;
					while (len > bufIndex)
					{
						kbLine[len] = kbLine[len-1];
						len--;
					}
					kbLine[bufIndex++] = c;

					index = bufIndex - 1;
					while (kbLine[index])
					{
						PutChar( kbLine[index++] );
					}
					while (index > bufIndex)
					{
						PutChar( '\b' );
						index--;
					}
					break;
			}
		break;

		case 1:
			if (c == '[')
			{
				escFlag = 2;
			}
			else
			{
				escFlag = 0;
			}
			return NULL;

		case 2:
			if (c == 'A')
			{
 				if (lineReadPtr == lineWritePtr)
				{
 					index = 0;
 					while ((lastLine[lineWritePtr][index] = kbLine[index])) index++;
				}

				do
				{
					if (--lineReadPtr < 0) lineReadPtr = MAX_LINES - 1;
					if (lastLine[lineReadPtr][0] != '\0') break;
				} while (lineReadPtr != lineWritePtr);

				while (bufIndex)						// delete the written line
				{
					PutChar( '\b' );
					PutChar( ' ' );
					PutChar( '\b' );
					bufIndex--;
				}

				while ((kbLine[bufIndex] = lastLine[lineReadPtr][bufIndex]))
				{
					PutChar( kbLine[bufIndex++] );
				}
			}
			else if (c == 'B')
			{
 				if (lineReadPtr == lineWritePtr)
				{
 					index = 0;
 					while ((lastLine[lineWritePtr][index] = kbLine[index])) index++;
				}

//				Printf( "lineReadPtr = %d; lineWritePtr = %d\n", lineReadPtr, lineWritePtr );
				do
				{
					if (++lineReadPtr >= MAX_LINES) lineReadPtr = 0;
					if (lastLine[lineReadPtr][0] != '\0') break;
				} while (lineReadPtr != lineWritePtr);

				while (bufIndex)						// delete the written line
				{
					PutChar( '\b' );
					PutChar( ' ' );
					PutChar( '\b' );
					bufIndex--;
				}

				while ((kbLine[bufIndex] = lastLine[lineReadPtr][bufIndex]))
				{
					PutChar( kbLine[bufIndex++] );
				}
			}
			else if (c == 'C')
			{
				if (bufIndex < len)
				{
					PutChar( kbLine[bufIndex++] );

				}
			}
			else if (c == 'D')
			{
				if (bufIndex)
				{
					PutChar( '\b' );
					bufIndex--;
				}
			}

			escFlag = 0;
			return NULL;

		default:
			escFlag = 0;
			break;
	}

	return NULL;
}
*/
